#
# Copyright (c) 2003 Gilles Lenfant
# All rights reserved. Written by Gilles Lenfant <glenfant@bigfoot.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the author may not be used to endorse or promote products
#    derived from this software without specific prior written permission
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""This is the EpozDocument Product for Zope
An EpozDocument is a content object with Epoz edition capability
$Id: EpozDocument.py,v 1.2 2004/02/22 17:17:26 gilles Exp $
"""

__version__ = "$Revision: 1.2 $"[11:-2]

import urlparse

from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.ZCatalog.CatalogPathAwareness import CatalogAware
import Persistence
from Globals import InitializeClass
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from AccessControl import ClassSecurityInfo, getSecurityManager
from AccessControl.Permissions import change_images_and_files, view, \
     access_contents_information, add_documents_images_and_files, \
     view_management_screens
from DateTime import DateTime
from webdav.Lockable import ResourceLockedError
from webdav.WriteLockInterface import WriteLockInterface
from ZTUtils import make_query, LazyFilter
from DocumentTemplate.sequence import sort

# Python 2.3.x backport
try:
    dummy = True
except NameError, e:
    True = 1
    False = 0

################
## Main class ##
################

class EpozDocument(CatalogAware, PropertyManager, SimpleItem):
    """An EpozDocument is a content object with Epoz edition capability
    """
    __implements__ = (WriteLockInterface,)

    meta_type = 'EpozDocument'

    _properties = (
        {'id': 'title', 'type': 'string', 'mode': 'w'},
        {'id': 'lastEditDate', 'type': 'date', 'mode': 'w'},
        {'id': 'EpozDocumentTemplate', 'type': 'string', 'mode': 'w'}
        )

    manage_options = (
        ({'label': 'Edit',
          'action': 'manage_editForm',
          'help': ('EpozDocument', 'editForm.stx')},
         {'label': 'View',
          'action': ''}
         )
        + PropertyManager.manage_options
        + ({'label': 'ZCatalog',
            'action': 'manage_editCatalogerForm'},)
        + SimpleItem.manage_options)

    security = ClassSecurityInfo()

    ##################
    ## User methods ##
    ##################

    def __init__(self, id, title=''):
        """Let's create it
        """
        self._internalVersion = 2 # Provision for future upgrades
        self.id = id
        self.title = title
        self.lastEditDate = DateTime()
        self.EpozDocumentTemplate = ''
        self.html = ''
        return

    # Prevent accidental acquisition
    index_html = None

    security.declarePublic('__call__')
    def __call__(self, client=None, REQUEST=None, RESPONSE=None, **kw):
        """Publishing (through the EpozDocumentTemplate if any)
        """
        if REQUEST is None:
            REQUEST = self.REQUEST
        
        templateId = 'EpozDocumentTemplate'
        templateId = getattr(self.aq_base, templateId) or templateId
        template = getattr(self.aq_parent, templateId, None)
        if template:
            template = template.__of__(self)
            return apply(template, ((client, self), REQUEST), kw)

        # Publish directly as HTML
        html = "<html><head><title>%s</title></head><body>%s</body></html>"
        html = html % (self.title, self.html)
        if RESPONSE is None:
            RESPONSE = REQUEST.RESPONSE
        RESPONSE.setHeader('Content-Type', 'text/html')
        RESPONSE.setHeader('Content-Length', str(len(html)))
        return html

    security.declareProtected(change_images_and_files, 'edit')
    def edit(self, html, title=None, EpozDocumentTemplate=None):
        """Changes the object content
        @param title: The new title
        @type title: string
        @param html: The new html content
        @type html: HTML text
        """
        if title is not None:
            self.title = title.strip()
        self.lastEditDate = DateTime()
        if EpozDocumentTemplate is not None:
            self.EpozDocumentTemplate = EpozDocumentTemplate.strip()
        # Runs the post edit hook if fonud
        hook = getattr(self, 'EpozDocumentPostEditHook', None)
        if hook is not None:
            html = hook(self, html)
        self.html = html.strip()
        self.reindex_object()
        return

    security.declarePublic('SearchableText')
    def SearchableText(self):
        """Provides all text for indexing
        """
        extrasId = 'SearchableTextAdd'
        textTypes = ('string', 'text')
        extraText = ''
        if self.hasProperty(extrasId):
            for extra in self.getProperty(extrasId):
                if (self.hasProperty(extra) and
                    (self.getPropertyType(extra) in textTypes)):
                    extraText += self.getProperty(extra) + ' '
        return " ".join((self.title, self.html, extraText))

    PrincipiaSearchSource = SearchableText


    #################
    ## ZMI methods ##
    #################

    manage_editForm = PageTemplateFile('zmi/manage_editForm', globals())
    manage_main = manage_editForm

    security.declareProtected(change_images_and_files, 'manage_edit')
    def manage_edit(self, title, EpozDocumentTemplate, html, REQUEST=None):
        """Edit the stuff (from ZMI)
        """
        if self.wl_isLocked():
            raise ResourceLockedError, "EpozDocument is locked via WebDAV"
        self.edit(html, title=title, EpozDocumentTemplate=EpozDocumentTemplate)
        if REQUEST:
            message="Saved changes."
            return self.manage_editForm(self,REQUEST,manage_tabs_message=message,
                                        management_view='Edit')
        return

    security.declareProtected(change_images_and_files, 'manage_widgetdims')
    def manage_widgetdims(self, REQUEST):
        """Change dimensions of widget and record dimensions in cookies
        """
        defaultWidth = 600
        defaultHeight = 300
        increment = 50
        change = REQUEST.form['manage_widgetdims']
        rules = {
            'Taller': ('__epozdocument_height__', defaultHeight, increment),
            'Shorter': ('__epozdocument_height__', defaultHeight, -increment),
            'Wider': ('__epozdocument_width__', defaultWidth, increment),
            'Narrower': ('__epozdocument_width__', defaultWidth, -increment)}
        cookieName, defaultSize, add = rules[change]
        expires = (DateTime('GMT') + 365).rfc822()
        size = REQUEST.get(cookieName, str(defaultSize))
        size = str(int(size) + add)
        # Makes change global for all EpozDocument objects of this Zope
        REQUEST.RESPONSE.setCookie(cookieName, size, path='/', expires=expires)
        message="Widget %s of %d pixels" % (change.lower(), increment)
        REQUEST.RESPONSE.redirect(REQUEST.URL1 + "/manage_editForm?"
                                  + make_query(manage_tabs_message=message,
                                               management_view='Edit'))
        return

    # Template for the images/links insertion
    epoz_toolbox = PageTemplateFile('zmi/epoz_toolbox', globals())

    security.declareProtected(change_images_and_files, 'epoz_toolbox_data_')
    def epoz_toolbox_data_(self, REQUEST):
        """A helper for the epoz_toolbox
        (Much easier to make data in Python than in ZPT :-)
        """
        # content_type & charset settings
        charset = REQUEST.get('management_page_charset')
        if not charset:
            charset = getattr(self, 'management_page_charset', 'iso-8859-1')
        REQUEST.set('management_page_charset', charset)
        content_type = "text/html;charset=%s" % charset
        REQUEST.RESPONSE.setHeader('Content-Type', content_type)

        # breadcrumb
        locationPath = REQUEST.get('toolbox_location', list(self.aq_parent.getPhysicalPath()))
        breadcrumb = []
        for i in range(len(locationPath)):
            path = locationPath[:i+1]
            breadcrumb.append(EpozToolBreadcrumbItem(REQUEST, path))

        # Target objects
        location = self.restrictedTraverse(locationPath)
        allObjects = LazyFilter(location.objectValues(), skip=view)
        sortOrder = (('getId', 'nocase', 'asc'),)

        # folderishs
        folderishs = LazyFilter(allObjects, test=keepFolderish, skip=access_contents_information)
        folderishs = sort(folderishs, sort=sortOrder)
        folderishs = [EpozToolFolderishTarget(self, o) for o in folderishs]

        # epozdocuments
        epozdocuments = LazyFilter(allObjects, test=keepEpozDocuments)
        epozdocuments = sort(epozdocuments, sort=sortOrder)
        epozdocuments = [EpozToolEpozDocumentTarget(self, o) for o in epozdocuments]

        # images
        images = LazyFilter(allObjects, test=keepImages)
        images = sort(images, sort=sortOrder)
        images = [EpozToolImageTarget(self, o) for o in images]

        # others
        others = LazyFilter(allObjects, test=keepOthers)
        others = sort(others, sort=sortOrder)
        others = [EpozToolOtherTarget(self, o) for o in others]
        
        # Making epoz_toolbox template data
        alldata = {
            'content_type': content_type,
            'locationpath': locationPath,
            'breadcrumb': breadcrumb,
            'folderishs': folderishs,
            'epozdocuments': epozdocuments,
            'images': images,
            'others': others
            }
        return alldata
    
    ####################
    ## Common methods ##
    ####################

    security.declarePublic('get_size')
    def get_size(self):
        """The size of the content
        """
        return len(self.html)

    # Support for deprecated method
    getSize = get_size

    ########################################
    ## FTP/WebDav/External Editor support ##
    ########################################

    security.declareProtected(change_images_and_files, 'PUT')
    def PUT(self, REQUEST, RESPONSE):
        """Handle HTTP PUT requests
        """
        self.dav__init(REQUEST, RESPONSE)
        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
        self.edit(REQUEST.get('BODY', ''))
        RESPONSE.setStatus(204)
        return RESPONSE

    security.declareProtected(change_images_and_files, 'manage_FTPget')
    def manage_FTPget(self):
        """Get source for FTP download
        """
        self.REQUEST.RESPONSE.setHeader('Content-Type', 'text/html')
        return self.html

    #####################
    ## Private methods ##
    #####################

    def _postInit(self):
        """Continues __init__(...) after setting in ZODB
        """
        self.index_object()
        return

    #################################################
    ## Upgrading EpozDocument object (if required) ##
    #################################################

    def __setstate__(self, state):
        """This is called before the EpozDocument is called and
        immediately after it is pulled out the ZODB.
        We use it for upgrading EpozDocument objects created with
        older versions.
        """
        Persistence.Persistent.__setstate__(self, state)

        if self._internalVersion < 2:
            self.lastEditDate = self.bobobase_modification_time()
            self._internalVersion = 2
        return

# /class EpozDocument

InitializeClass(EpozDocument)

#########################################
## Helper classes for the epoz_toolbox ##
#########################################

class EpozToolBreadcrumbItem:
    """Item for a breadcrumb
    """

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, request, path):
        """Constructor
        @param request: the published REQUEST
        @param path: targetted folderish object
        """
        self.request = request
        self.path = path
        return

    def label(self):
        """Displayed label
        """
        label = self.path[-1]
        if not label:
            label = "<i>Zope root</i>"
        elif not self._browsableTarget():
            label = "<i>Unauthorized</i>"
        return label

    def link(self):
        """URL target of link
        """
        if not self._browsableTarget():
            return ""
        queryString = make_query(toolbox_location=list(self.path))
        return self.request.URL + '?' + queryString

    def _browsableTarget(self):
        """Check if object may be browsed by the user
        (i.e in case of not root Manager)
        """
        try:
            return self._isBrowsable
        except AttributeError, e:
            target = '/'.join(self.path) or '/'
            target = self.request['PUBLISHED'].restrictedTraverse(target)
            user = getSecurityManager().getUser()
            self._isBrowsable = user.has_permission(view_management_screens, target)
            return self._isBrowsable
# /class EpozToolBreadcrumbItem


class EpozToolTargetBase:
    """Base class for iarget item in epoz tool
    """
    
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, reference, objekt):
        """Constructor
        @param reference: This EpozDocument object
        @param objekt: A potential target object
        """
        self.reference = reference
        self.objekt = objekt
        return

    def icon(self):
        """Transfer to obj attribute"""
        return self.objekt.icon

    def absolute_url(self):
        """Transfer to obj attribute"""
        return self.objekt.absolute_url()

    def meta_type(self):
        """Transfer to obj attribute"""
        return self.objekt.meta_type

    def getId(self):
        """Transfer to obj attribute"""
        return self.objekt.getId()

    def title(self):
        """Transfer to obj attribute"""
        return self.objekt.title
    
    def relative_url(self):
        """Relative URL from reference to objekt
        suitable to a "href" or "src" attribute
        """
        # urlparse components
        SCHEME, HOSTNAME, PATH, PARAMS, QUERY, FRAGMENT = range(6)
        orgUnits = urlparse.urlparse(self.reference.absolute_url())
        targUnits = urlparse.urlparse(self.objekt.absolute_url())
        afterPath = targUnits[PARAMS:]
        if orgUnits[:PATH] != targUnits[:PATH]:
            # Scheme (http) or host:port (www.comain.com) are different
            # -> no possible relative URL
            return targetUrl
        orgPath = orgUnits[PATH].split('/')
        orgPath.reverse()
        targPath = targUnits[PATH].split('/')
        targPath.reverse()

        # Remove parts which are equal (['a', 'b'] ['a', 'c'] --> ['c'])
        while (len(orgPath) > 0
               and len(targPath) > 0
               and orgPath[-1] == targPath[-1]):
            orgPath.pop()
            last_pop = targPath.pop()
        if len(orgPath) == 0 and len(targPath) == 0:
            # Special case: link to itself (http://foo/a http://foo/a -> a)
            targPath.append(last_pop)
        if len(orgPath) == 1 and orgPath[0] == "" and len(targPath) == 0:
            # Special case: (http://foo/a/ http://foo/a -> ../a)
            orgPath.append(last_pop)
            targPath.append(last_pop)
        targPath.reverse()
        relativeUrl = ['..'] * (len(orgPath) -1)
        relativeUrl = ('', '', '/'.join(relativeUrl + targPath)) + tuple(afterPath)
        return urlparse.urlunparse(relativeUrl)
# /class EpozToolTargetBase


class EpozToolFolderishTarget(EpozToolTargetBase):
    """Folderish obj in the epox_toolbox
    """
    def browseurl(self):
        """URL for browsing this folderish
        """
        baseUrl = self.reference.absolute_url() + '/epoz_toolbox'
        queryStr = make_query(toolbox_location=list(self.objekt.getPhysicalPath()))
        return baseUrl + '?' + queryStr

class EpozToolEpozDocumentTarget(EpozToolTargetBase):
    """Other EpozDocument content
    """
    pass

class EpozToolImageTarget(EpozToolTargetBase):
    """Image obj in the epoz_toolbox
    """
    def width(self):
        """Transfer to obj attribute"""
        return self.objekt.width

    def height(self):
        """Transfer to obj attribute"""
        return self.objekt.height

class EpozToolOtherTarget(EpozToolTargetBase):
    """Any other obj in the epoz_toolbox
    """
    pass

############################
## Filters for LazyFilter ##
############################

def keepFolderish(objekt):
    """Function for LazyFilter
    """
    return objekt.isPrincipiaFolderish

def keepEpozDocuments(objekt):
    """Function for LazyFilter
    """
    return objekt.meta_type == 'EpozDocument'

def keepImages(objekt):
    """Function for LazyFilter
    """
    imageMetaTypes = ('Image',) # Maybe others
    return objekt.meta_type in imageMetaTypes

def keepOthers(objekt):
    """Function for LazyFilter
    keeps those not filtered by above filters
    """
    filters = (keepFolderish, keepEpozDocuments, keepImages)
    for filter in filters:
        if filter(objekt):
            return False
    return True

###############################
## Zope factory registration ##
###############################

def initialize(context):
    """Registers this Product into Zope
    """
    context.registerClass(
        instance_class = EpozDocument,
        permission=add_documents_images_and_files,
        constructors=(manage_addEpozDocumentForm, manage_addEpozDocument),
        icon='www/EpozDocument_icon.gif')
    context.registerHelp()
    return

########################
## Zope factory hooks ##
########################

def manage_addEpozDocument(container, id, title='', REQUEST=None):
    """Creates an EpozDocument in the container
    """
    id = id.strip()
    o = EpozDocument(id, title=title)
    container._setObject(id, o)
    o._postInit()
    if REQUEST:
        if REQUEST.form.has_key("add"):
            nextPath = '/manage_workspace'
        else:
            # "addedit" button clicked
            nextPath = '/%s/manage_workspace' % id
        REQUEST['RESPONSE'].redirect(container.absolute_url() + nextPath)
    return

manage_addEpozDocumentForm = PageTemplateFile('zmi/addEpozDocumentForm', globals())
